import logging
from uuid import uuid4

from sourced.ml.extractors import IdentifiersBagExtractor
from sourced.ml.transformers import Cacher, UastDeserializer, CooccConstructor, CooccModelSaver, \
    Uast2BagFeatures, Counter, UastRow2Document, Repartitioner, create_uast_source
from sourced.ml.utils.engine import pipeline_graph, pause
from sourced.ml.utils.docfreq import create_or_load_ordered_df


@pause
def repos2coocc(args):
    log = logging.getLogger("repos2coocc")
    id_extractor = IdentifiersBagExtractor(docfreq_threshold=args.min_docfreq,
                                           split_stem=args.split)
    session_name = "repos2coocc-%s" % uuid4()
    root, start_point = create_uast_source(args, session_name)

    uast_extractor = start_point \
        .link(UastRow2Document()) \
        .link(Repartitioner.maybe(args.partitions, args.shuffle)) \
        .link(Cacher.maybe(args.persist))
    log.info("Extracting UASTs...")
    ndocs = uast_extractor.link(Counter()).execute()
    log.info("Number of documents: %d", ndocs)
    uast_extractor = uast_extractor.link(UastDeserializer())

    df_model = create_or_load_ordered_df(
        args, ndocs, uast_extractor.link(Uast2BagFeatures([id_extractor])))

    token2index = root.session.sparkContext.broadcast(df_model.order)
    uast_extractor \
        .link(CooccConstructor(token2index=token2index,
                               token_parser=id_extractor.id2bag.token_parser,
                               namespace=id_extractor.NAMESPACE)) \
        .link(CooccModelSaver(args.output, df_model)) \
        .execute()
    pipeline_graph(args, log, root)



from io import StringIO

from sourced.ml.utils import PickleableLogger  # nopep8


class Transformer(PickleableLogger):
    BLOCKS = False  # set to True if __call__ launches PySpark

    def __init__(self, explain=None, **kwargs):
        super().__init__(**kwargs)
        self._children = []
        self._parent = None
        self._explained = explain

    def __getstate__(self):
        state = super().__getstate__()
        del state["_parent"]
        del state["_children"]
        return state

    @property
    def explained(self):
        if self._explained is None and self.parent is not None:
            return self.parent.explained
        return bool(self._explained)

    @property
    def children(self):
        return tuple(self._children)

    @property
    def parent(self):
        return self._parent

    def path(self):
        node = self
        path = []
        while node is not None:
            path.append(node)
            node = node.parent
        path.reverse()
        return path

    def link(self, transformer):
        self._children.append(transformer)
        transformer._parent = self
        return transformer

    def unlink(self, transformer):
        self._children.remove(transformer)
        transformer._parent = None
        return self

    def _explode(self, head, context):
        if context[-1] is not self:
            context.append(self)
            if not self._children or self.BLOCKS:
                self._log.info(self._format_pipeline(context))
            head = self(head)
        results = []
        for child in self._children:
            results.extend(child._explode(head, context.copy()))
        else:
            results.append(head)
        return results

    def explode(self, head=None):
        """
        Execute all the branches in the tree going from the node. The node itself
        is execute()-ed.
        :param head: The input to feed to the root. Can be None - the default one \
                     will be used if possible.
        :return: The results from all the leaves.
        """
        head = self.execute(head)
        pipeline = [self]
        node = self
        while node.parent is not None:
            node = node.parent
            pipeline.append(node)
        pipeline.reverse()
        return self._explode(head, pipeline)

    def execute(self, head=None):
        """
        Execute the node together with all its dependencies, in order.
        :param head: The input to feed to the ultimate parent. Can be None - the default one \
                     will be used if possible.
        :return: The result of the execution.
        """
        pipeline = self.path()
        if not self._children:
            self._log.info(self._format_pipeline(pipeline))
        for node in pipeline:
            head = node(head)
        return head

    def graph(self, name="source-d", stream=None):
        if stream is None:
            stream = StringIO()
        stream.write("digraph %s {\n" % name)
        counters = {}
        nodes = {}
        queue = [(self, None)]
        while queue:
            node, parent = queue.pop(0)
            try:
                myself = nodes[node]
            except KeyError:
                index = counters.setdefault(type(node), 0)
                index += 1
                counters[type(node)] = index
                myself = "%s %d" % (type(node).__name__, index)
                nodes[node] = myself
            if parent is not None:
                stream.write("\t\"%s\" -> \"%s\"\n" % (parent, myself))
            for child in node._children:
                queue.append((child, myself))
        stream.write("}\n")
        return stream

    def _get_log_name(self):
        return self.__class__.__name__

    @staticmethod
    def _format_pipeline(pipeline):
        return " -> ".join(type(n).__name__ for n in pipeline)

    def __call__(self, head):
        raise NotImplementedError()



from sourced.ml.algorithms.uast_ids_to_bag import UastIds2Bag
from sourced.ml.extractors import BagsExtractor, register_extractor
from sourced.ml.algorithms import NoopTokenParser


@register_extractor
class IdentifiersBagExtractor(BagsExtractor):
    NAME = "id"
    NAMESPACE = "i."
    OPTS = {"split-stem": True}
    OPTS.update(BagsExtractor.OPTS)

    def __init__(self, docfreq_threshold=None, split_stem=True, **kwargs):
        super().__init__(docfreq_threshold, **kwargs)
        self.id2bag = UastIds2Bag(None, NoopTokenParser() if not split_stem else None)

    def uast_to_bag(self, uast):
        return self.id2bag(uast)



from datetime import datetime
import logging
import random
import os
from typing import Callable, Iterable, List, Tuple
import warnings

import numpy
import keras
from keras import backend as kbackend
from keras.callbacks import CSVLogger, TensorBoard, ModelCheckpoint, LearningRateScheduler
try:
    import tensorflow as tf
except ImportError:
    warnings.warn("Tensorflow is not installed, dependent functionality is unavailable.")


# additional variable to avoid any division by zero when computing the precision and recall metrics
EPSILON = 10 ** -8
# threshold that is used to binarize predictions of the model
DEFAULT_THRESHOLD = 0.5


def set_random_seed(seed: int) -> None:
    """
    Fixes a random seed for reproducibility.

    :param seed: seed value.
    """
    numpy.random.seed(seed)
    random.seed(seed)
    tf.set_random_seed(seed)


def binarize(matrix: numpy.array, threshold: float, inplace: bool=True) -> numpy.array:
    """
    Helper function to binarize a matrix.

    :param matrix: matrix as a numpy.array.
    :param threshold: if value >= threshold then the value will be 1, else 0.
    :param inplace: whether to modify the matrix inplace or not.
    :return: the binarized matrix.
    """
    mask = matrix >= threshold
    if inplace:
        matrix_ = matrix
    else:
        matrix_ = matrix.copy()
    matrix_[mask] = 1
    matrix_[numpy.logical_not(mask)] = 0
    return matrix_


def str2ints(params: str) -> List[int]:
    """
    Convert a string with integer parameters to a list of integers.

    :param params: string that contains integer parameters separated by commas.
    :return: list of integers.
    """
    return list(map(int, params.split(",")))


def precision_np(y_true: numpy.array, y_pred: numpy.array, epsilon: float=EPSILON) -> float:
    """
    Computes the precision metric, a metric for multi-label classification of
    how many selected items are relevant.

    :param y_true: ground truth labels - expect binary values.
    :param y_pred: predicted labels - expect binary values.
    :param epsilon: added to the denominator to avoid any division by zero.
    :return: precision metric.
    """
    true_positives = numpy.sum(y_true * y_pred)
    predicted_positives = numpy.sum(y_pred)
    return true_positives / (predicted_positives + epsilon)


def recall_np(y_true: numpy.array, y_pred: numpy.array, epsilon: float=EPSILON) -> float:
    """
    Computes the recall metric, a metric for multi-label classification of
    how many relevant items are selected.

    :param y_true: matrix with ground truth labels - expect binary values.
    :param y_pred: matrix with predicted labels - expect binary values.
    :param epsilon: added to the denominator to avoid any division by zero.
    :return: recall metric.
    """
    true_positives = numpy.sum(y_true * y_pred)
    possible_positives = numpy.sum(y_true)
    return true_positives / (possible_positives + epsilon)


def report(model: keras.engine.training.Model, X: numpy.array, y: numpy.array, batch_size: int,
           threshold: float=DEFAULT_THRESHOLD, epsilon: float=EPSILON) -> None:
    """
    Prints a metric report of the `model` on the  data `X` & `y`.
    The metrics printed are precision, recall, F1 score.

    :param model: model considered.
    :param X: features.
    :param y: labels (expected binary labels).
    :param batch_size: batch size that will be used for prediction.
    :param threshold: threshold to binarize the predictions.
    :param epsilon: added to the denominator to avoid any division by zero.
    """
    log = logging.getLogger("report")

    # predict & skip the last dimension & binarize
    predictions = model.predict(X, batch_size=batch_size, verbose=1)[:, :, 0]
    predictions = binarize(predictions, threshold)

    # report
    pr = precision_np(y[:, :, 0], predictions, epsilon=epsilon)
    rec = recall_np(y[:, :, 0], predictions, epsilon=epsilon)
    f1 = 2 * pr * rec / (pr + rec + epsilon)
    log.info("precision: %.3f, recall: %.3f, f1: %.3f" % (pr, rec, f1))


def config_keras() -> None:
    """
    Initializes keras backend session.
    """
    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    kbackend.tensorflow_backend.set_session(tf.Session(config=config))


def build_train_generator(X: numpy.array, y: numpy.array,
                          batch_size: int=500) -> Iterable[Tuple[numpy.array]]:
    """
    Builds the generator that yields features and their labels.

    :param X: features.
    :param y: binary labels.
    :param batch_size: higher values better utilize GPUs.
    :return: generator of features and their labels.
    """
    assert X.shape[0] == y.shape[0], "Number of samples mismatch in X and y."

    def xy_generator():
        while True:
            n_batches = X.shape[0] // batch_size
            if n_batches * batch_size < X.shape[0]:
                n_batches += 1  # to yield last samples
            for i in range(n_batches):
                start = i * batch_size
                end = min((i + 1) * batch_size, X.shape[0])
                yield X[start:end], y[start:end]
    return xy_generator()


def build_schedule(lr: float, final_lr: float, n_epochs: int) -> Callable:
    """
    Builds the schedule of which the learning rate decreases.
    The schedule makes the learning rate decrease linearly.

    :param lr: initial learning rate.
    :param final_lr: final learning rate.
    :param n_epochs: number of training epochs.
    :return: the schedule of the learning rate.
    """
    delta = (lr - final_lr) / n_epochs

    def schedule(epoch: int) -> float:
        assert 0 <= epoch < n_epochs
        return lr - delta * epoch
    return schedule


def make_lr_scheduler(lr: float, final_lr: float, n_epochs: int,
                      verbose: int=1) -> keras.callbacks.LearningRateScheduler:
    """
    Prepares the scheduler to decrease the learning rate while training.

    :param lr: initial learning rate.
    :param final_lr: final learning rate.
    :param n_epochs: number of training epochs.
    :param verbose: level of verbosity.
    :return: LearningRateScheduler with linear schedule of the learning rate.
    """
    schedule = build_schedule(lr, final_lr, n_epochs)
    return LearningRateScheduler(schedule=schedule, verbose=verbose)


def prepare_callbacks(output_dir: str) -> Tuple[Callable]:
    """
    Prepares logging, tensorboard, model checkpoint callbacks and stores the outputs in output_dir.

    :param output_dir: path to the results.
    :return: list of callbacks.
    """
    time = datetime.now().strftime("%y%m%d-%H%M")
    log_dir = os.path.join(output_dir, "tensorboard" + time)
    logging.info("Tensorboard directory: %s" % log_dir)
    tensorboard = TensorBoard(log_dir=log_dir, batch_size=1000, write_images=True,
                              write_graph=True)
    csv_path = os.path.join(output_dir, "csv_logger_" + time + ".txt")
    logging.info("CSV logs: %s" % csv_path)
    csv_logger = CSVLogger(csv_path)

    filepath = os.path.join(output_dir, "best_" + time + ".model")
    model_saver = ModelCheckpoint(filepath, monitor='val_recall', verbose=1, save_best_only=True,
                                  mode='max')
    return tensorboard, csv_logger, model_saver


def create_generator_params(batch_size: int, samples_per_epoch: int, n_samples: int,
                            epochs: int) -> Tuple[int]:
    """
    Helper function to split a huge dataset into smaller ones to enable more frequent reports.

    :param batch_size: batch size.
    :param samples_per_epoch: number of samples per mini-epoch or before each report.
    :param n_samples: total number of samples.
    :param epochs: number of epochs over the full dataset.
    :return: number of steps per epoch (should be used with the generator) and number of sub-epochs
             where during sub-epoch only samples_per_epoch will be generated.
    """
    steps_per_epoch = samples_per_epoch // batch_size
    n_epochs = numpy.ceil(epochs * n_samples / samples_per_epoch)
    return steps_per_epoch, n_epochs



import logging

from sourced.ml.transformers import BagFeatures2DocFreq, Uast2BagFeatures
from sourced.ml.models import OrderedDocumentFrequencies


def create_or_load_ordered_df(args, ndocs: int, bag_features: Uast2BagFeatures):
    """
    Returns a preexisting OrderedDocumentFrequencies model from docfreq_in, or generates one
    from the flattened bags of features using args and saves it to docfreq_out.

    :param args: Instance of `argparse.Namespace` that contains docfreq_in, docfreq_out,
                 min_docfreq, and vocabulary_size.
    :param ndocs: Number of documents (can be repos, files or functions)
    :param bag_features: Transformer containing bags of features extracted from the data (the call
                         instantiates an RDD: [(key, doc), val] where key is a specific feature
                         that appeared val times in the document doc.
    :return: OrderedDocumentFrequencies model
    """
    log = logging.getLogger("create_or_load_ordered_df")
    if args.docfreq_in:
        log.info("Loading ordered docfreq model from %s ...", args.docfreq_in)
        return OrderedDocumentFrequencies().load(args.docfreq_in)
    log.info("Calculating the document frequencies, hold tight ...")
    df = bag_features \
        .link(BagFeatures2DocFreq()) \
        .execute()
    log.info("Writing ordered docfreq model to %s ...", args.docfreq_out)
    df_model = OrderedDocumentFrequencies() \
        .construct(ndocs, df) \
        .prune(args.min_docfreq) \
        .greatest(args.vocabulary_size) \
        .save(args.docfreq_out)
    return df_model



import logging
import multiprocessing
import os
from typing import Union, List

from modelforge import Model
from modelforge.progress_bar import progress_bar

from sourced.ml.utils.pickleable_logger import PickleableLogger


class Model2Base(PickleableLogger):
    """
    Base class for model -> model conversions.
    """
    MODEL_FROM_CLASS = None
    MODEL_TO_CLASS = None

    def __init__(self, num_processes: int=0,
                 log_level: int=logging.DEBUG, overwrite_existing: bool=True):
        """
        Initializes a new instance of Model2Base class.

        :param num_processes: The number of processes to execute for conversion.
        :param log_level: Logging verbosity level.
        :param overwrite_existing: Rewrite existing models or skip them.
        """
        super().__init__(log_level=log_level)
        self.num_processes = multiprocessing.cpu_count() if num_processes == 0 else num_processes
        self.overwrite_existing = overwrite_existing

    def convert(self, models_path: List[str], destdir: str, pattern: str="**/*.asdf") -> int:
        """
        Performs the model -> model conversion. Runs the conversions in a pool of processes.

        :param srcdir: List of Models path.
        :param destdir: The directory where to store the models. The directory structure is \
                        preserved.
        :param pattern: glob pattern for the files.
        :return: The number of converted files.
        """
        files = list(models_path)
        self._log.info("Found %d files", len(files))
        if not files:
            return 0
        queue_in = multiprocessing.Manager().Queue()
        queue_out = multiprocessing.Manager().Queue(1)
        processes = [multiprocessing.Process(target=self._process_entry,
                                             args=(i, destdir, queue_in, queue_out))
                     for i in range(self.num_processes)]
        for p in processes:
            p.start()
        for f in files:
            queue_in.put(f)
        for _ in processes:
            queue_in.put(None)
        failures = 0
        for _ in progress_bar(files, self._log, expected_size=len(files)):
            filename, ok = queue_out.get()
            if not ok:
                failures += 1
        for p in processes:
            p.join()
        self._log.info("Finished, %d failed files", failures)
        return len(files) - failures

    def convert_model(self, model: Model) -> Union[Model, None]:
        """
        This must be implemented in the child classes.

        :param model: The model instance to convert.
        :return: The converted model instance or None if it is not needed.
        """
        raise NotImplementedError

    def finalize(self, index: int, destdir: str):
        """
        Called for each worker in the end of the processing.

        :param index: Worker's index.
        :param destdir: The directory where to store the models.
        """
        pass

    def _process_entry(self, index, destdir, queue_in, queue_out):
        while True:
            filepath = queue_in.get()
            if filepath is None:
                break
            try:
                model_path = os.path.join(destdir, os.path.split(filepath)[1])
                if os.path.exists(model_path):
                    if self.overwrite_existing:
                        self._log.warning(
                            "Model %s already exists, but will be overwrite. If you want to "
                            "skip existing models use --disable-overwrite flag", model_path)
                    else:
                        self._log.warning("Model %s already exists, skipping.", model_path)
                        queue_out.put((filepath, True))
                        continue
                model_from = self.MODEL_FROM_CLASS(log_level=self._log.level).load(filepath)
                model_to = self.convert_model(model_from)
                if model_to is not None:
                    dirs = os.path.dirname(model_path)
                    if dirs:
                        os.makedirs(dirs, exist_ok=True)
                    model_to.save(model_path, deps=model_to.meta["dependencies"])
            except:  # noqa
                self._log.exception("%s failed", filepath)
                queue_out.put((filepath, False))
            else:
                queue_out.put((filepath, True))
        self.finalize(index, destdir)

    def _get_log_name(self):
        return "%s2%s" % (self.MODEL_FROM_CLASS.NAME, self.MODEL_TO_CLASS.NAME)

    def _get_model_path(self, path):
        """
        By default, we name the converted files exactly the same.

        :param path: The path relative to ``srcdir``.
        :return: The target path for the converted model.
        """
        return path



import operator
import os
from typing import NamedTuple, Generator
import yaml

import pygments
from pygments.formatter import Formatter
from pygments.lexers import get_lexer_by_name, ClassNotFound
from pyspark import RDD, Row
from pyspark.sql import DataFrame, functions

from sourced.ml.algorithms import TokenParser
from sourced.ml.transformers import Transformer
from sourced.ml.utils import EngineConstants


class ContentToIdentifiers(Transformer):

    class FormatterProxy(Formatter):
        name = "Proxy"
        aliases = ["proxy"]
        filenames = []

        def __init__(self, **options):
            super(ContentToIdentifiers.FormatterProxy, self).__init__(**options)
            self.callback = options["callback"]

        def format(self, tokensource, outfile):
            self.callback(tokensource)

    def __init__(self, split: bool, **kwargs):
        super().__init__(**kwargs)
        Columns = NamedTuple("Columns", [("repo_id", str), ("file_id", str)])
        self.column_names = Columns(repo_id=EngineConstants.Columns.RepositoryId,
                                    file_id=EngineConstants.Columns.Path)
        self.linguist2pygments = self.build_mapping()
        self.split = split

    def __call__(self, rows: DataFrame):
        rdd = rows .where(functions.length(functions.col("content")) > 0).rdd
        return rdd.flatMap(self.process_row)

    def process_row(self, row: Row):
        self.names = []
        repo_id = getattr(row, self.column_names.repo_id)
        file_id = getattr(row, self.column_names.file_id)
        path = os.path.join(repo_id, file_id)
        code = row.content
        for i in (0, 1):
            try:
                lexer = get_lexer_by_name(self.linguist2pygments[row.lang][i])
                pygments.highlight(code, lexer, self.FormatterProxy(callback=self.process_tokens))
                break
            except (KeyError, IndexError, ClassNotFound, UnicodeDecodeError) as e:
                continue
        for token in self.names:
            yield token, (repo_id, path)

    def process_tokens(self, tokens: Generator):
        """
        Filter tokens of type "Name" and which are splittable
        according to :class: 'TokenParser' rules
        """
        for _type, token in tokens:
            if _type[0] == "Name":
                if self.split:
                    if sum(1 for _ in TokenParser(min_split_length=1).split(token)) > 1:
                        self.names.append(token)
                else:
                    self.names.append(token)

    @staticmethod
    def build_mapping():
        """
        Builds the mapping between linguist languages and pygments names for lexers.
        """
        linguist2pygments = {}
        with open(os.path.join(os.path.dirname(__file__), "languages.yml")) as f:
            all_languages = yaml.load(f)

        linguist_langs = {}
        for lang, specs in all_languages.items():
            if specs["type"] == "programming":
                linguist_langs[lang] = (set(specs.get("aliases", []) + [lang.lower()]), specs)

        pygments_langs = set()
        for lexer in pygments.lexers.LEXERS.values():
            lang_declensions = [lang.lower() for lang in lexer[2]] + [lexer[1].lower()]
            pygments_langs |= set(lang_declensions)

        for lang in linguist_langs:
            lang_names = linguist_langs.get(lang, (set(),))[0]
            inter = list(lang_names.intersection(pygments_langs))
            if inter:
                linguist2pygments[lang] = inter
        return linguist2pygments


class IdentifiersToDataset(Transformer):

    def __init__(self, idfreq: bool, **kwargs):
        super().__init__(**kwargs)
        self.idfreq = idfreq

    def __call__(self, rows: RDD):
        if self.idfreq:
            return self.process_stats(rows)
        else:
            return self.process(rows)

    def reduce_rows(self, rows: RDD):
        return rows \
            .map(lambda x: (x[0], 1)) \
            .reduceByKey(operator.add)

    def process_stats(self, rows: RDD):
        """
        Process rows to gather identifier frequencies.
        num_repos is the number of repositories where the identifier appears in.
        num_files is the number of files where the identifier appears in.
        num_occ is the total number of occurences of the identifier.
        """
        list_RDDs = []
        for i in (0, 1):
            # initial structure of x: (identifier, (repositoryId, filepath))
            freq_processed = rows.map(lambda x: (x[0], x[1][i])).distinct()
            list_RDDs.append(self.reduce_rows(freq_processed))
        list_RDDs.append(self.reduce_rows(rows))
        return rows \
            .context.union(list_RDDs) \
            .groupByKey() \
            .mapValues(list) \
            .map(lambda x: Row(
                    token=x[0],
                    token_split=" ".join(TokenParser(min_split_length=1).split(x[0])),
                    num_repos=x[1][0],
                    num_files=x[1][1],
                    num_occ=x[1][2]))

    def process(self, rows: RDD):
        """
        Process rows to gather identifier without frequencies.
        """
        return rows \
            .keys() \
            .distinct() \
            .map(lambda x: Row(
                    token=x,
                    token_split=" ".join(TokenParser(min_split_length=1).split(x))))



from sourced.ml.utils import PickleableLogger


class BagsExtractor(PickleableLogger):
    """
    Converts a single UAST into the weighted set (dictionary), where elements are strings
    and the values are floats. The derived classes must implement uast_to_bag().
    """
    DEFAULT_DOCFREQ_THRESHOLD = 5
    NAMESPACE = None  # the beginning of each element in the bag
    NAME = None  # feature scheme name, should be overridden in the derived class.
    OPTS = {"weight": 1}  # cmdline args which are passed into __init__()

    def __init__(self, docfreq_threshold=None, weight=None, **kwargs):
        """
        :param docfreq_threshold: The minimum number of occurrences of an element to be included \
                                  into the bag
        :param weight: TF-IDF will be multiplied by this weight to change importance of specific \
                      bag extractor
        """
        super().__init__(**kwargs)
        if docfreq_threshold is None:
            docfreq_threshold = self.DEFAULT_DOCFREQ_THRESHOLD
        self.docfreq_threshold = docfreq_threshold
        self.docfreq = {}
        self._ndocs = 0
        if weight is None:
            self.weight = 1
        else:
            self.weight = weight

    @property
    def docfreq_threhold(self):
        return self._docfreq_threshold

    @docfreq_threhold.setter
    def docfreq_threshold(self, value):
        if not isinstance(value, int):
            raise TypeError("docfreq_threshold must be an integer, got %s" % type(value))
        if value < 1:
            raise ValueError("docfreq_threshold must be >= 1, got %d" % value)
        self._docfreq_threshold = value

    @property
    def ndocs(self):
        return self._ndocs

    @ndocs.setter
    def ndocs(self, value):
        if not isinstance(value, int):
            raise TypeError("ndocs must be an integer, got %s" % type(value))
        if value < 1:
            raise ValueError("ndocs must be >= 1, got %d" % value)
        self._ndocs = value

    def _get_log_name(self):
        return type(self).__name__

    def extract(self, uast):
        for key, val in self.uast_to_bag(uast).items():
            yield self.NAMESPACE + key, val * self.weight

    @classmethod
    def get_kwargs_fromcmdline(cls, args):
        prefix = cls.NAME + "_"
        result = {}
        for k, v in args.__dict__.items():
            if k.startswith(prefix):
                result[k[len(prefix):]] = v
        return result

    def uast_to_bag(self, uast):
        raise NotImplemented



from pyspark import Row

from sourced.ml.algorithms import log_tf_log_idf
from sourced.ml.models import DocumentFrequencies
from sourced.ml.transformers.uast2bag_features import Uast2BagFeatures
from sourced.ml.transformers.transformer import Transformer


class TFIDF(Transformer):
    """
    Calculates TF-IDF (log-log) weight for every feature.
    """
    Columns = Uast2BagFeatures.Columns

    def __init__(self, df: DocumentFrequencies, **kwargs):
        """
        :param tf: pyspark rdd where each row is named tuple with `token`, `document` and `value` \
                   (term frequency) fields. One can use Uast2TermFreq Transformer to calculate \
                   such rdd.
        :param df: DocumentFrequencies model calculated for the given `tf` stream.
        """
        super().__init__(**kwargs)
        self.df = df

    def __call__(self, head):
        c = self.Columns
        df = self.df
        return head \
            .filter(lambda x: df.get(x[c.token]) is not None) \
            .map(lambda x: Row(**{
                c.token: x[c.token],
                c.document: x[c.document],
                c.value: log_tf_log_idf(df=df[x[c.token]], tf=x[c.value], ndocs=df.docs)}))



# flake8: noqa
from sourced.ml.utils.bigartm import install_bigartm
from sourced.ml.utils.bblfsh_roles import IDENTIFIER, QUALIFIED, LITERAL, FUNCTION, DECLARATION, \
    NAME
from sourced.ml.utils.spark import add_spark_args, create_spark, get_spark_memory_config, \
    SparkDefault
from sourced.ml.utils.engine import add_engine_args, create_engine, EngineConstants, EngineDefault
from sourced.ml.utils.pickleable_logger import PickleableLogger



import logging
import os
import sys
import pkg_resources
import zipfile

os.environ["PYSPARK_PYTHON"] = sys.executable

import pyspark  # noqa
from pyspark.sql import SparkSession  # noqa


class SparkDefault:
    """
    Default arguments for create_spark function and __main__
    """
    MASTER_ADDRESS = "local[*]"
    LOCAL_DIR = "/tmp/spark"
    LOG_LEVEL = "WARN"
    CONFIG = (
        # to clean up unpacked Siva files, see https://github.com/src-d/engine/issues/348
        "spark.tech.sourced.engine.cleanup.skip=false",
        # to skip broken siva files
        "spark.tech.sourced.engine.skip.read.errors=true",
    )
    JAR_PACKAGES = tuple()
    MEMORY = ""
    STORAGE_LEVEL = "MEMORY_AND_DISK"
    DEP_ZIP = False


def add_spark_args(my_parser, default_packages=None):
    my_parser.add_argument(
        "-s", "--spark", default=SparkDefault.MASTER_ADDRESS,
        help="Spark's master address.")
    my_parser.add_argument(
        "--config", nargs="+", default=SparkDefault.CONFIG,
        help="Spark configuration (key=value).")
    my_parser.add_argument(
        "-m", "--memory", default=SparkDefault.MEMORY,
        help="Handy memory config for spark. -m 4G,10G,2G is equivalent to "
             "--config spark.executor.memory=4G "
             "--config spark.driver.memory=10G "
             "--config spark.driver.maxResultSize=2G."
             "Numbers are floats separated by commas.")
    if default_packages is None:
        default_packages = SparkDefault.JAR_PACKAGES
    my_parser.add_argument(
        "--package", nargs="+", default=default_packages, dest="packages",
        help="Additional Spark packages.")
    my_parser.add_argument(
        "--spark-local-dir", default=SparkDefault.LOCAL_DIR,
        help="Spark local directory.")
    spark_log_levels = ("ALL", "DEBUG", "ERROR", "FATAL", "INFO", "OFF", "TRACE", "WARN")
    my_parser.add_argument("--spark-log-level", default=SparkDefault.LOG_LEVEL,
                           choices=spark_log_levels, help="Spark log level")
    my_parser.add_argument("--dep-zip", default=SparkDefault.DEP_ZIP,
                           action="store_true", help="Adds ml and engine to the spark context.")
    persistences = [att for att in pyspark.StorageLevel.__dict__.keys() if "__" not in att]
    my_parser.add_argument(
        "--persist", default=SparkDefault.STORAGE_LEVEL, choices=persistences,
        help="Spark persistence type (StorageLevel.*).")
    my_parser.add_argument(
        "--pause", action="store_true",
        help="Do not terminate in the end - useful for inspecting Spark Web UI.")
    my_parser.add_argument("--explain", action="store_true",
                           help="Print the PySpark execution plans.")


def create_spark(session_name,
                 spark=SparkDefault.MASTER_ADDRESS,
                 spark_local_dir=SparkDefault.LOCAL_DIR,
                 config=SparkDefault.CONFIG,
                 packages=SparkDefault.JAR_PACKAGES,
                 spark_log_level=SparkDefault.LOG_LEVEL,
                 memory=SparkDefault.MEMORY,
                 dep_zip=SparkDefault.DEP_ZIP):
    log = logging.getLogger("spark")
    log.info("Starting %s on %s", session_name, spark)
    # Hide py4j verbose logging (It appears in travis mostly)
    logging.getLogger("py4j").setLevel(logging.WARNING)
    config += get_spark_memory_config(memory)
    builder = SparkSession.builder.master(spark).appName(session_name)
    builder = builder.config("spark.jars.packages", ",".join(packages))
    builder = builder.config("spark.local.dir", spark_local_dir)
    for cfg in config:
        builder = builder.config(*cfg.split("=", 1))
    session = builder.getOrCreate()
    session.sparkContext.setLogLevel(spark_log_level)
    if dep_zip:
        zip_path = zip_sourced()
        session.sparkContext.addPyFile(zip_path)
    return session


def zip_sourced():
    zip_path = os.path.expanduser("~/.cache/sourced/ml/sourced.zip")
    if not os.path.exists(zip_path):
        os.makedirs(zip_path.split("sourced.zip")[0])
        working_dir = os.getcwd()
        os.chdir(pkg_resources.working_set.by_key["sourced-ml"].location)
        with zipfile.PyZipFile(zip_path, "w", zipfile.ZIP_DEFLATED) as zf:
            zf.write("sourced")
            zf.writepy("sourced/ml", "sourced")
            zf.writepy("sourced/engine", "sourced")
        os.chdir(working_dir)
    return zip_path


def get_spark_memory_config(memory=SparkDefault.MEMORY):
    """
    Assemble memory configuration for a Spark session
    :param memory: string with memory configuration for spark
    :return: memory configuration
    """
    if not memory:
        return tuple()
    memory = memory.split(",")
    if len(memory) != 3:
        raise ValueError("Expected 3 memory parameters but got %s. Please check --help "
                         "to see how -m/--memory should be used." % len(memory))
    return ("spark.executor.memory=" + memory[0],
            "spark.driver.memory=" + memory[1],
            "spark.driver.maxResultSize=" + memory[2])



from typing import Iterable

import bblfsh

from sourced.ml.algorithms import Uast2IdSequence
from sourced.ml.extractors.bags_extractor import BagsExtractor
from sourced.ml.algorithms import NoopTokenParser


class IdSequenceExtractor(BagsExtractor):
    """
    Extractor wrapper for Uast2RoleIdPairs algorithm.
    Note that this is unusual BagsExtractor since it returns iterable instead of bag.

    The class did not wrap with @register_extractor because it does not produce bags as others do.
    So nobody outside code will see it or use it directly.
    For the same reason we a free to override NAMESPACE, NAME, OPTS fields with any value we want.

    TODO(zurk): Split BagsExtractor into two clases: Extractor and BagsExtractor(Extractor),
    re-inherit this class from Extractor, delete explanations from docstring.
    """
    NAMESPACE = ""
    NAME = "id sequence"
    OPTS = {}

    def __init__(self, split_stem=False, **kwargs):
        super().__init__(**kwargs)
        self.uast2id_sequence = Uast2IdSequence(
            None, NoopTokenParser() if not split_stem else None)

    def extract(self, uast: bblfsh.Node) -> Iterable[str]:
        yield self.uast2id_sequence(uast), None



from typing import Iterable, Union

from pyspark import RDD, Row
from pyspark.sql import DataFrame

from sourced.ml.extractors import BagsExtractor
from sourced.ml.transformers.transformer import Transformer
from sourced.ml.utils import EngineConstants


class UastRow2Document(Transformer):
    REPO_PATH_SEP = "//"
    PATH_BLOB_SEP = "@"

    def __call__(self, rows: Union[RDD, DataFrame]):
        if isinstance(rows, DataFrame):
            rows = rows.rdd
        return rows.map(self.documentize)

    def documentize(self, r: Row) -> Row:
        ec = EngineConstants.Columns
        doc = r[ec.RepositoryId]
        if r[ec.Path]:
            doc += self.REPO_PATH_SEP + r[ec.Path]
        if r[ec.BlobId]:
            doc += self.PATH_BLOB_SEP + r[ec.BlobId]
        bfc = Uast2BagFeatures.Columns
        return Row(**{bfc.document: doc, ec.Uast: r[ec.Uast]})


class Uast2BagFeatures(Transformer):
    class Columns:
        """
        Standard column names for interop.
        """
        token = "token"
        document = "document"
        value = "value"

    def __init__(self, extractors: Iterable[BagsExtractor], **kwargs):
        super().__init__(**kwargs)
        self.extractors = extractors

    def __call__(self, rows: RDD):
        return rows.flatMap(self.process_row)

    def process_row(self, row: Row):
        uast_column = EngineConstants.Columns.Uast
        doc = row[self.Columns.document]
        for uast in row[uast_column]:
            for extractor in self.extractors:
                for key, val in extractor.extract(uast):
                    yield (key, doc), val



from typing import Iterable, Tuple

import bblfsh

from sourced.ml.extractors.bags_extractor import BagsExtractor
from sourced.ml.algorithms import NoopTokenParser, Uast2IdTreeDistance, Uast2IdLineDistance
from sourced.ml.algorithms.uast_id_distance import Uast2IdDistance


class IdentifierDistance(BagsExtractor):
    """
    Extractor wrapper for Uast2IdTreeDistance and Uast2IdLineDistance algorithm.
    Note that this is an unusual BagsExtractor since it returns iterable instead of bag.

    The class did not wrap with @register_extractor because it does not produce bags as others do.
    So nobody outside code will see it or use it directly.
    For the same reason we a free to override NAMESPACE, NAME, OPTS fields with any value we want.

    TODO(zurk): Split BagsExtractor into two clases: Extractor and BagsExtractor(Extractor),
    re-inherit this class from Extractor, delete explanations from docstring.
    """
    NAMESPACE = ""
    NAME = "Identifier distance"
    OPTS = {}
    DEFAULT_MAX_DISTANCE = Uast2IdDistance.DEFAULT_MAX_DISTANCE

    class DistanceType:
        Tree = "tree"
        Line = "line"
        All = {Tree, Line}

        @staticmethod
        def resolve(type):
            if type == IdentifierDistance.DistanceType.Line:
                return Uast2IdLineDistance
            if type == IdentifierDistance.DistanceType.Tree:
                return Uast2IdTreeDistance
            raise ValueError("Unknown distance type: %s" % type)

    def __init__(self, split_stem=False, type="tree", max_distance=DEFAULT_MAX_DISTANCE, **kwargs):
        super().__init__(**kwargs)
        Uast2IdDistance = self.DistanceType.resolve(type)
        self.uast2id_distance = Uast2IdDistance(
            token_parser=NoopTokenParser() if not split_stem else None,
            max_distance=max_distance)

    def extract(self, uast: bblfsh.Node) -> Iterable[Tuple[str, str, int]]:
        yield from self.uast2id_distance(uast)



from typing import List

from modelforge import register_model, Model
import numpy


@register_model
class TensorFlowModel(Model):
    """
    TensorFlow Protobuf model exported in the Modelforge format with GraphDef inside.
    """
    NAME = "tensorflow-model"

    def construct(self, graphdef: "tensorflow.GraphDef"=None, session: "tensorflow.Session"=None,
                  outputs: List[str]=None):
        if graphdef is None:
            assert session is not None
            assert outputs is not None
            graphdef = session.graph_def
            from tensorflow.python.framework import graph_util
            for node in graphdef.node:
                node.device = ""
                graphdef = graph_util.convert_variables_to_constants(
                    session, graphdef, outputs)
        self._graphdef = graphdef
        return self

    @property
    def graphdef(self):
        """
        Returns the wrapped TensorFlow GraphDef.
        """
        return self._graphdef

    def _generate_tree(self) -> dict:
        return {"graphdef": numpy.frombuffer(self._graphdef.SerializeToString(),
                                             dtype=numpy.uint8)}

    def _load_tree(self, tree: dict):
        from tensorflow.core.framework import graph_pb2

        graphdef = graph_pb2.GraphDef()
        graphdef.ParseFromString(tree["graphdef"].data)
        self.construct(graphdef=graphdef)



import os
import logging
from typing import List

from sourced.ml.models import QuantizationLevels
from sourced.ml.transformers import Uast2Quant
from sourced.ml.extractors import BagsExtractor


def create_or_apply_quant(model_path: str, extractors: List[BagsExtractor], extracted_uasts=None):
    log = logging.getLogger("create_or_apply_quant")
    if os.path.exists(model_path):
        QuantizationLevels().load(source=model_path).apply_quantization(extractors)
    else:
        quant = Uast2Quant(extractors)
        extracted_uasts.link(quant).execute()
        if quant.levels:
            log.info("Writing quantization levels to %s", model_path)
            QuantizationLevels().construct(quant.levels).save(model_path)



from typing import Iterable, Tuple

import bblfsh

from sourced.ml.algorithms.uast_ids_to_bag import UastIds2Bag
from sourced.ml.utils import bblfsh_roles


class Uast2RoleIdPairs(UastIds2Bag):
    """
    Converts a UAST to a list of pairs. Pair is identifier and role, where role is Node role
    where identifier was found.

    __call__ is overridden here and returns list instead of bag-of-words (dist).
    """

    def __init__(self, token2index=None, token_parser=None):
        """
        :param token2index: The mapping from tokens to token key. If None, no mapping is performed.
        :param token_parser: Specify token parser if you want to use a custom one. \
            :class:'TokenParser' is used if it is not specified.

        """
        super().__init__(token2index=token2index, token_parser=token_parser)
        self.exclude_roles = {
            bblfsh_roles.EXPRESSION,
            bblfsh_roles.IDENTIFIER,
            bblfsh_roles.LEFT,
            bblfsh_roles.QUALIFIED,
            bblfsh_roles.BINARY,
            bblfsh_roles.ASSIGNMENT,
        }

    def __call__(self, uast: bblfsh.Node) -> Iterable[Tuple[str, str]]:
        """
        Converts a UAST to a list of identifier, role pairs.
        The tokens are preprocessed by _token_parser.

        :param uast: The UAST root node.
        :return: a list of identifier, role pairs.
        """
        yield from self._process_uast(uast, [])

    def _process_uast(self, uast: bblfsh.Node, ancestors):
        stack = [(uast, [])]
        while stack:
            node, ancestors = stack.pop()

            if bblfsh_roles.IDENTIFIER in node.roles and node.token:
                roles = set(node.roles)
                indx = -1
                # We skip all Nodes with roles from `self.exclude_roles` set.
                # We skip any Node with OPERATOR role.
                # For them we take first parent Node from stack with another Role set.
                while not (roles - self.exclude_roles and bblfsh_roles.OPERATOR not in roles):
                    roles = set(ancestors[indx].roles)
                    indx -= 1
                for sub in self._token_parser.process_token(node.token):
                    try:
                        yield (self._token2index[sub], self.merge_roles(roles))
                    except KeyError:
                        continue
            ancestors = list(ancestors)
            ancestors.append(node)
            stack.extend([(child, ancestors) for child in node.children])

    @staticmethod
    def merge_roles(roles: Iterable[int]):
        return " | ".join(bblfsh.role_name(r) for r in sorted(roles))



import logging
from uuid import uuid4

from sourced.ml.extractors import create_extractors_from_args
from sourced.ml.transformers import UastDeserializer, BagFeatures2TermFreq, Uast2BagFeatures, \
    HeadFiles, TFIDF, Cacher, Indexer, UastRow2Document, BOWWriter, Moder, create_uast_source, \
    Repartitioner
from sourced.ml.utils.engine import pipeline_graph, pause
from sourced.ml.utils.docfreq import create_or_load_ordered_df
from sourced.ml.utils.quant import create_or_apply_quant


@pause
def repos2bow_template(args, select=HeadFiles, cache_hook=None, save_hook=None):
    log = logging.getLogger("repos2bow")
    extractors = create_extractors_from_args(args)
    session_name = "repos2bow-%s" % uuid4()
    root, start_point = create_uast_source(args, session_name, select=select)
    uast_extractor = start_point.link(Moder(args.mode)) \
        .link(Repartitioner.maybe(args.partitions, args.shuffle)) \
        .link(Cacher.maybe(args.persist))
    if cache_hook is not None:
        uast_extractor.link(cache_hook()).execute()
    # We link UastRow2Document after Cacher here because cache_hook() may want to have all possible
    # Row items.
    uast_extractor = uast_extractor.link(UastRow2Document())
    log.info("Extracting UASTs and indexing documents...")
    document_indexer = Indexer(Uast2BagFeatures.Columns.document,
                               cached_index_path=args.cached_index_path)
    if args.cached_index_path is None:
        uast_extractor.link(document_indexer).execute()
    ndocs = len(document_indexer)
    log.info("Number of documents: %d", ndocs)
    uast_extractor = uast_extractor.link(UastDeserializer())
    if args.quant:
        create_or_apply_quant(args.quant, extractors, uast_extractor)
    uast_extractor = uast_extractor \
        .link(Uast2BagFeatures(extractors))
    df_model = create_or_load_ordered_df(args, ndocs, uast_extractor)
    bags_writer = uast_extractor \
        .link(BagFeatures2TermFreq()) \
        .link(TFIDF(df_model)) \
        .link(document_indexer) \
        .link(Indexer(Uast2BagFeatures.Columns.token, df_model.order))
    if save_hook is not None:
        bags_writer = bags_writer \
            .link(Repartitioner.maybe(args.partitions, args.shuffle, multiplier=10)) \
            .link(save_hook())
    bags_writer.link(BOWWriter(document_indexer, df_model, args.bow, args.batch)) \
        .execute()
    pipeline_graph(args, log, root)


def repos2bow(args):
    return repos2bow_template(args)


def repos2bow_index_template(args, select=HeadFiles):
    log = logging.getLogger("repos2bow_index")
    extractors = create_extractors_from_args(args)
    session_name = "repos2bow_index_features-%s" % uuid4()
    root, start_point = create_uast_source(args, session_name, select=select)
    uast_extractor = start_point.link(Moder(args.mode)) \
        .link(Repartitioner.maybe(args.partitions, args.shuffle)) \
        .link(UastRow2Document()) \
        .link(Cacher.maybe(args.persist))
    log.info("Extracting UASTs and indexing documents...")
    document_indexer = Indexer(Uast2BagFeatures.Columns.document)
    uast_extractor.link(document_indexer).execute()
    document_indexer.save_index(args.cached_index_path)
    ndocs = len(document_indexer)
    log.info("Number of documents: %d", ndocs)
    uast_extractor = uast_extractor.link(UastDeserializer())
    if args.quant:
        create_or_apply_quant(args.quant, extractors, uast_extractor)
    if args.docfreq_out:
        create_or_load_ordered_df(args, ndocs, uast_extractor.link(Uast2BagFeatures(extractors)))
    pipeline_graph(args, log, root)


def repos2bow_index(args):
    return repos2bow_index_template(args)



import logging
from uuid import uuid4

from sourced.ml.extractors import create_extractors_from_args
from sourced.ml.models import OrderedDocumentFrequencies
from sourced.ml.transformers import UastDeserializer, BagFeatures2DocFreq, Uast2BagFeatures, \
    Counter, Cacher, UastRow2Document, create_uast_source
from sourced.ml.utils.engine import pipeline_graph, pause
from sourced.ml.utils.quant import create_or_apply_quant


@pause
def repos2df(args):
    log = logging.getLogger("repos2df")
    extractors = create_extractors_from_args(args)
    session_name = "repos2df-%s" % uuid4()
    root, start_point = create_uast_source(args, session_name)

    uast_extractor = start_point \
        .link(UastRow2Document()) \
        .link(Cacher.maybe(args.persist))
    log.info("Extracting UASTs...")
    ndocs = uast_extractor.link(Counter()).execute()
    log.info("Number of documents: %d", ndocs)
    uast_extractor = uast_extractor.link(UastDeserializer())
    if args.quant:
        create_or_apply_quant(args.quant, extractors, uast_extractor)
    df = uast_extractor \
        .link(Uast2BagFeatures(extractors)) \
        .link(BagFeatures2DocFreq()) \
        .execute()
    log.info("Writing docfreq model to %s", args.docfreq_out)
    OrderedDocumentFrequencies().construct(ndocs, df).save(args.docfreq_out)
    pipeline_graph(args, log, root)



import operator

from pyspark import Row, RDD

from sourced.ml.transformers.uast2bag_features import Uast2BagFeatures
from sourced.ml.transformers.transformer import Transformer


class BagFeatures2TermFreq(Transformer):
    def __call__(self, rows: RDD):
        c = Uast2BagFeatures.Columns
        return rows \
            .reduceByKey(operator.add) \
            .map(lambda x: Row(**{
                c.token: x[0][0], c.document: x[0][1], c.value: x[1],
            }))



import argparse
import logging
import os
import pickle

from sourced.ml.algorithms.id_splitter import set_random_seed, config_keras, \
    prepare_features, create_generator_params, str2ints, build_train_generator, \
    prepare_devices, build_rnn, build_cnn, prepare_callbacks, make_lr_scheduler, \
    report


def train_id_split(args: argparse.ArgumentParser):
    """
    Pipeline to train a neural network to split identifiers.

    :param args: arguments.
    :param model: type of neural network used to learn the splitting task.
    """
    log = logging.getLogger("train_id_split")
    config_keras()
    set_random_seed(args.seed)

    # prepare features
    X_train, X_test, y_train, y_test = prepare_features(
        csv_path=args.input,
        use_header=args.include_csv_header,
        max_identifier_len=args.length,
        identifier_col=args.csv_identifier,
        split_identifier_col=args.csv_identifier_split,
        test_ratio=args.test_ratio,
        padding=args.padding)

    # prepare train generator
    steps_per_epoch, n_epochs = create_generator_params(
        batch_size=args.batch_size,
        samples_per_epoch=args.samples_before_report,
        n_samples=X_train.shape[0],
        epochs=args.epochs)

    train_gen = build_train_generator(X=X_train, y=y_train, batch_size=args.batch_size)

    # prepare test generator
    validation_steps, _ = create_generator_params(batch_size=args.val_batch_size,
                                                  samples_per_epoch=X_test.shape[0],
                                                  n_samples=X_test.shape[0],
                                                  epochs=args.epochs)
    test_gen = build_train_generator(X=X_test, y=y_test, batch_size=args.val_batch_size)

    # initialize model
    dev0, dev1 = prepare_devices(args.devices)

    # build a RNN model
    if args.model == "RNN":
        model = build_rnn(maxlen=args.length,
                          units=args.neurons,
                          stack=args.stack,
                          optimizer=args.optimizer,
                          rnn_layer=args.type_cell,
                          dev0=dev0,
                          dev1=dev1)

    else:
        # build a CNN model
        filters = str2ints(args.filters)
        kernel_sizes = str2ints(args.kernel_sizes)
        model = build_cnn(maxlen=args.length,
                          filters=filters,
                          output_n_filters=args.dim_reduction,
                          stack=args.stack,
                          kernel_sizes=kernel_sizes,
                          optimizer=args.optimizer,
                          device=dev0)

    log.info("Model summary:")
    model.summary(print_fn=log.info)

    # callbacks
    callbacks = prepare_callbacks(args.output)
    lr_scheduler = make_lr_scheduler(lr=args.lr, final_lr=args.final_lr, n_epochs=n_epochs)
    callbacks = callbacks + (lr_scheduler,)

    # train
    history = model.fit_generator(train_gen, steps_per_epoch=steps_per_epoch,
                                  validation_data=test_gen, validation_steps=validation_steps,
                                  callbacks=list(callbacks), epochs=n_epochs)

    # report quality on test dataset
    report(model, X=X_test, y=y_test, batch_size=args.val_batch_size)

    # save model & history
    # TODO: Use modelforge to save the model
    with open(os.path.join(args.output, "model_history.pickle"), "wb") as f:
        pickle.dump(history.history, f)
    model.save(os.path.join(args.output, "last.model"))
    log.info("Completed!")



import logging
import string
import tarfile
from typing import List, Tuple

import numpy
from keras.preprocessing.sequence import pad_sequences
from modelforge.progress_bar import progress_bar


def read_identifiers(csv_path: str, use_header: bool, max_identifier_len: int, identifier_col: int,
                     split_identifier_col: int, shuffle: bool=True) -> List[str]:
    """
    Reads and filters too long identifiers in the CSV file.

    :param csv_path: path to the CSV file.
    :param use_header: uses header as normal line (True) or treat as header line with column names.
    :param max_identifier_len: maximum length of raw identifiers. Skip identifiers that are longer.
    :param identifier_col: column name in the CSV file for the raw identifier.
    :param split_identifier_col: column name in the CSV file for the splitted identifier lowercase.
    :param shuffle: indicates whether to reorder the list of identifiers
        at random after reading it.
    :return: list of splitted identifiers.
    """
    log = logging.getLogger("read_identifiers")
    log.info("Reading data from the CSV file %s", csv_path)
    identifiers = []
    # TODO: Update dataset loading as soon as https://github.com/src-d/backlog/issues/1212 done
    # Think about dataset download step
    with tarfile.open(csv_path, encoding="utf-8") as f:
        assert len(f.members) == 1, "One archived file is expected, got: %s" % len(f.members)
        content = f.extractfile(f.members[0])
        if not use_header:
            content.readline()
        for line in progress_bar(content.readlines(), log):
            row = line.decode("utf-8").strip().split(",")
            if len(row[identifier_col]) <= max_identifier_len:
                identifiers.append(row[split_identifier_col])
    if shuffle:
        numpy.random.shuffle(identifiers)
    log.info("Number of identifiers after filtering: %s." % len(identifiers))
    return identifiers


def prepare_features(csv_path: str, use_header: bool, max_identifier_len: int,
                     identifier_col: int, split_identifier_col: int, test_ratio: float,
                     padding: str, shuffle: bool=True) -> Tuple[numpy.array]:
    """
    Prepare the features to train the identifier splitting task.

    :param csv_path: path to the CSV file.
    :param use_header: uses header as normal line (True) or treat as header line with column names.
    :param max_identifier_len: maximum length of raw identifiers. Skip identifiers that are longer.
    :param identifier_col: column in the CSV file for the raw identifier.
    :param split_identifier_col: column in the CSV file for the splitted identifier.
    :param shuffle: indicates whether to reorder the list of identifiers
        at random after reading it.
    :param test_ratio: Proportion of test samples used for evaluation.
    :param padding: position where to add padding values:
        after the intput sequence if "post", before if "pre".
    :return: training and testing features to train the neural net for the splitting task.
    """
    log = logging.getLogger("prepare_features")

    # read data from the input file
    identifiers = read_identifiers(csv_path=csv_path, use_header=use_header,
                                   max_identifier_len=max_identifier_len,
                                   identifier_col=identifier_col,
                                   split_identifier_col=split_identifier_col, shuffle=shuffle)

    # convert identifiers into character indices and labels
    log.info("Converting identifiers to character indices")
    log.info("Number of identifiers: %d, Average length: %d characters" %
             (len(identifiers), numpy.mean([len(i) for i in identifiers])))

    char2ind = dict((c, i + 1) for i, c in enumerate(sorted(string.ascii_lowercase)))

    char_id_seq = []
    splits = []
    for identifier in identifiers:
        # iterate through the identifier and convert to array of char indices & boolean split array
        index_arr = []
        split_arr = []
        skip_char = False
        for char in identifier.strip():
            if char in char2ind:
                index_arr.append(char2ind[char])
                if skip_char:
                    skip_char = False
                    continue
                split_arr.append(0)
            elif char == " ":
                split_arr.append(1)
                skip_char = True
            else:
                log.warning("Unexpected symbol %s in identifier", char)
        # sanity check
        assert len(index_arr) == len(split_arr)
        char_id_seq.append(index_arr)
        splits.append(split_arr)

    log.info("Number of subtokens: %d, Number of distinct characters: %d" %
             (sum([sum(split_arr) for split_arr in splits]) + len(identifiers),
              len(set([i for index_arr in char_id_seq for i in index_arr]))))

    # train/test splitting
    log.info("Train/test splitting...")
    n_train = int((1 - test_ratio) * len(char_id_seq))
    X_train = char_id_seq[:n_train]
    X_test = char_id_seq[n_train:]
    y_train = splits[:n_train]
    y_test = splits[n_train:]
    log.info("Number of train samples: %s, number of test samples: %s" % (len(X_train),
                                                                          len(X_test)))

    # pad sequence
    log.info("Padding the sequences...")
    X_train = pad_sequences(X_train, maxlen=max_identifier_len, padding=padding)
    X_test = pad_sequences(X_test, maxlen=max_identifier_len, padding=padding)
    y_train = pad_sequences(y_train, maxlen=max_identifier_len, padding=padding)
    y_test = pad_sequences(y_test, maxlen=max_identifier_len, padding=padding)

    return X_train, X_test, y_train[:, :, None], y_test[:, :, None]



from typing import Iterable

import bblfsh

from sourced.ml.algorithms.uast_id_distance import Uast2IdLineDistance


class Uast2IdSequence(Uast2IdLineDistance):
    """
    Converts a UAST to a sorted sequence of identifiers.
    Identifiers are sorted by position in code.
    We do not change the order if positions are not present.

    __call__ is overridden here and return list instead of bag-of-words (dist).
    """

    def __call__(self, uast: bblfsh.Node) -> str:
        """
        Converts a UAST to a sorted sequence of identifiers.
        Identifiers are sorted by position in code.
        We do not change the order if positions are not present.

        :param uast: The UAST root node.
        :return: string with a sequence of identifiers
        """
        return self.concat(id for id, pos in sorted(self._process_uast(uast), key=lambda x: x[1]))

    @staticmethod
    def concat(id_sequence: Iterable):
        return ' '.join(id_sequence)



# flake8: noqa
from sourced.ml.cmd.args import ArgumentDefaultsHelpFormatterNoNone
from sourced.ml.cmd.bigartm2asdf import bigartm2asdf
from sourced.ml.cmd.bow_converters import bow2vw
from sourced.ml.cmd.merge_df import merge_df
from sourced.ml.cmd.merge_coocc import merge_coocc
from sourced.ml.cmd.id2vec_postprocess import id2vec_postprocess
from sourced.ml.cmd.id2vec_preprocess import id2vec_preprocess
from sourced.ml.cmd.preprocess_repos import preprocess_repos
from sourced.ml.cmd.id2vec_project import id2vec_project
from sourced.ml.cmd.repos2bow import repos2bow, repos2bow_template, repos2bow_index, \
    repos2bow_index_template
from sourced.ml.cmd.repos2coocc import repos2coocc
from sourced.ml.cmd.repos2df import repos2df
from sourced.ml.cmd.repos2ids import repos2ids
from sourced.ml.cmd.train_id_split import train_id_split
from sourced.ml.cmd.run_swivel import run_swivel
from sourced.ml.cmd.repos2roles_and_ids import repos2roles_and_ids
from sourced.ml.cmd.repos2id_distance import repos2id_distance
from sourced.ml.cmd.repos2id_sequence import repos2id_sequence



import os
import logging
from uuid import uuid4

from sourced.ml.transformers import FieldsSelector, Moder, DzhigurdaFiles, ParquetSaver, \
    create_uast_source
from sourced.ml.utils.engine import pipeline_graph, pause


@pause
def preprocess_repos(args):
    log = logging.getLogger("preprocess_repos")
    session_name = "preprocess_repos-%s" % uuid4()

    if os.path.exists(args.output):
        log.critical("%s must not exist", args.output)
        return 1
    if not args.config:
        args.config = []
    root, start_point = create_uast_source(args, session_name,
                                           select=lambda: DzhigurdaFiles(args.dzhigurda))

    start_point \
        .link(Moder(args.mode)) \
        .link(FieldsSelector(fields=args.fields)) \
        .link(ParquetSaver(save_loc=args.output)) \
        .execute()
    pipeline_graph(args, log, root)



import logging
from uuid import uuid4

from sourced.ml.extractors.roles_and_ids import RolesAndIdsExtractor
from sourced.ml.transformers import UastDeserializer, Uast2BagFeatures, UastRow2Document, \
    CsvSaver, create_uast_source
from sourced.ml.transformers.basic import Rower
from sourced.ml.utils.engine import pipeline_graph, pause


@pause
def repos2roles_and_ids(args):
    log = logging.getLogger("repos2roles_and_ids")
    session_name = "repos2roles_and_ids-%s" % uuid4()
    root, start_point = create_uast_source(args, session_name)

    start_point \
        .link(UastRow2Document()) \
        .link(UastDeserializer()) \
        .link(Uast2BagFeatures([RolesAndIdsExtractor(args.split)])) \
        .link(Rower(lambda x: dict(identifier=x[0][0], role=x[1]))) \
        .link(CsvSaver(args.output)) \
        .execute()
    pipeline_graph(args, log, root)



import logging
from uuid import uuid4

from sourced.ml.transformers import ContentToIdentifiers, create_uast_source, LanguageSelector, \
    IdentifiersToDataset, CsvSaver, Repartitioner
from sourced.ml.utils.engine import pipeline_graph, pause


@pause
def repos2ids(args):
    log = logging.getLogger("repos2ids")
    session_name = "repos2ids-%s" % uuid4()
    language_selector = LanguageSelector(languages=["null"], blacklist=True)
    root, start_point = create_uast_source(args, session_name, language_selector=language_selector,
                                           extract_uast=False)
    start_point \
        .link(Repartitioner(args.partitions, args.shuffle)) \
        .link(ContentToIdentifiers(args.split)) \
        .link(IdentifiersToDataset(args.idfreq)) \
        .link(CsvSaver(args.output)) \
        .execute()
    pipeline_graph(args, log, root)



import logging
from typing import Union

from pyspark import RDD, Row, StorageLevel
from pyspark.sql import DataFrame, functions

from sourced.ml.extractors.helpers import filter_kwargs
from sourced.ml.transformers.transformer import Transformer
from sourced.ml.transformers.uast2bag_features import Uast2BagFeatures
from sourced.ml.utils import EngineConstants, get_spark_memory_config, \
    create_engine, create_spark, SparkDefault


class Repartitioner(Transformer):
    def __init__(self, partitions: int, shuffle: bool=False, **kwargs):
        super().__init__(**kwargs)
        self.partitions = partitions
        self.shuffle = shuffle

    def __call__(self, head: RDD):
        return head.coalesce(self.partitions, self.shuffle)

    @staticmethod
    def maybe(partitions: Union[int, None], shuffle: bool=False, multiplier: int=1):
        if partitions is not None:
            return Repartitioner(partitions * multiplier, shuffle)
        else:
            return Identity()


class CsvSaver(Transformer):
    def __init__(self, output: str, **kwargs):
        super().__init__(**kwargs)
        self.output = output

    def __call__(self, head: RDD):
        self._log.info("Writing %s", self.output)
        return head.toDF() \
            .coalesce(1) \
            .write \
            .option("header", "true") \
            .mode("overwrite") \
            .csv(self.output)


class Rower(Transformer):
    def __init__(self, dicter, **kwargs):
        super().__init__(**kwargs)
        self.dicter = dicter

    def __call__(self, head: RDD):
        return head.map(lambda x: Row(**self.dicter(x)))


class Sampler(Transformer):
    """
    Wraps `sample()` function from pyspark Dataframe.
    """
    def __init__(self, with_replacement=False, fraction=0.05, seed=42, **kwargs):
        super().__init__(**kwargs)
        self.with_replacement = with_replacement
        self.fraction = fraction
        self.seed = seed

    def __call__(self, head: RDD):
        return head.sample(self.with_replacement, self.fraction, self.seed)


class Collector(Transformer):
    def __call__(self, head: RDD):
        return head.collect()


class First(Transformer):
    def __call__(self, head: RDD):
        return head.first()


class Identity(Transformer):
    def __call__(self, head: RDD):
        return head


class Cacher(Transformer):
    def __init__(self, persistence, **kwargs):
        super().__init__(**kwargs)
        self.persistence = getattr(StorageLevel, persistence)
        self.head = None
        self.trace = None

    def __getstate__(self):
        state = super().__getstate__()
        state["head"] = None
        state["trace"] = None
        return state

    def __call__(self, head: RDD):
        if self.head is None or self.trace != self.path():
            self.head = head.persist(self.persistence)
            self.trace = self.path()
        return self.head

    @staticmethod
    def maybe(persistence):
        if persistence is not None:
            return Cacher(persistence)
        else:
            return Identity()


class Ignition(Transformer):
    def __init__(self, engine, **kwargs):
        super().__init__(**kwargs)
        self.engine = engine

    def __getstate__(self):
        state = super().__getstate__()
        del state["engine"]
        return state

    def __call__(self, _):
        return self.engine


class DzhigurdaFiles(Transformer):
    def __init__(self, dzhigurda, **kwargs):
        super().__init__(**kwargs)
        self.dzhigurda = dzhigurda

    def __call__(self, engine):
        head_ref = engine.repositories.references.head_ref
        if self.dzhigurda < 0:
            # Use all available commits
            chosen = head_ref.all_reference_commits
        elif self.dzhigurda == 0:
            # Use only the first commit on a reference.
            chosen = head_ref.commits
        else:
            commits = head_ref.all_reference_commits
            chosen = commits.filter(commits.index <= self.dzhigurda)
        return chosen.tree_entries.blobs


class HeadFiles(Transformer):
    def __call__(self, engine):
        return engine.repositories.references.head_ref.commits.tree_entries.blobs


class Counter(Transformer):
    def __init__(self, distinct=False, approximate=False, **kwargs):
        super().__init__(**kwargs)
        self.distinct = distinct
        self.approximate = approximate

    def __call__(self, head: RDD):
        if self.distinct and not self.approximate:
            head = head.distinct()
        if self.explained:
            self._log.info("toDebugString():\n%s", head.toDebugString().decode())
        if not self.approximate or not self.distinct:
            return head.count()
        return head.countApproxDistinct()


class LanguageSelector(Transformer):
    def __init__(self, languages: Union[list, tuple], blacklist=False, **kwargs):
        super().__init__(**kwargs)
        self.languages = languages
        self.blacklist = blacklist

    def __call__(self, files: DataFrame) -> DataFrame:
        files = files.dropDuplicates(("blob_id",)).filter("is_binary = 'false'")
        classified = files.classify_languages()
        if not self.blacklist:
            lang_filter = classified.lang == self.languages[0]
            for lang in self.languages[1:]:
                lang_filter |= classified.lang == lang
        else:
            lang_filter = classified.lang != self.languages[0]
            for lang in self.languages[1:]:
                lang_filter &= classified.lang != lang
        filtered_by_lang = classified.filter(lang_filter)
        return filtered_by_lang


class UastExtractor(Transformer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __call__(self, files: DataFrame) -> DataFrame:
        # if UAST is not extracted, returns an empty list that we filter out here
        return files.extract_uasts().where(functions.size(functions.col("uast")) > 0)


class FieldsSelector(Transformer):
    def __init__(self, fields: Union[list, tuple], **kwargs):
        super().__init__(**kwargs)
        self.fields = fields

    def __call__(self, rdd: RDD) -> RDD:
        def select_fields(row):
            return Row(**{f: getattr(row, f) for f in self.fields})
        res = rdd.map(select_fields)
        if self.explained:
            self._log.info("toDebugString():\n%s", res.toDebugString().decode())
        return res


class ParquetSaver(Transformer):
    def __init__(self, save_loc, **kwargs):
        super().__init__(**kwargs)
        self.save_loc = save_loc

    def __call__(self, rdd: RDD):
        if self.explained:
            self._log.info("toDebugString():\n%s", rdd.toDebugString().decode())
        rdd.toDF().write.parquet(self.save_loc)


class ParquetLoader(Transformer):
    def __init__(self, session, paths, **kwargs):
        super().__init__(**kwargs)
        self.session = session
        self.paths = paths

    def __getstate__(self):
        state = super().__getstate__()
        del state["session"]
        return state

    def __call__(self, _):
        if isinstance(self.paths, (list, tuple)):
            return self.session.read.parquet(*self.paths)
        if isinstance(self.paths, str):
            return self.session.read.parquet(self.paths)
        raise ValueError


class UastDeserializer(Transformer):
    def __setstate__(self, state):
        super().__setstate__(state)
        from bblfsh import Node
        self.parse_uast = Node.FromString

    def __call__(self, rows: RDD) -> RDD:
        return rows.flatMap(self.deserialize_uast)

    def deserialize_uast(self, row: Row):
        if not row[EngineConstants.Columns.Uast]:
            return
        row_dict = row.asDict()
        row_dict[EngineConstants.Columns.Uast] = []
        for i, uast in enumerate(row[EngineConstants.Columns.Uast]):
            try:
                row_dict[EngineConstants.Columns.Uast].append(self.parse_uast(uast))
            except:  # noqa
                self._log.error("\nBabelfish Error: Failed to parse uast for document %s for uast "
                                "#%s" % (row[Uast2BagFeatures.Columns.document], i))
        yield Row(**row_dict)


def create_parquet_loader(session_name, repositories,
                          config=SparkDefault.CONFIG,
                          packages=SparkDefault.JAR_PACKAGES,
                          spark=SparkDefault.MASTER_ADDRESS,
                          spark_local_dir=SparkDefault.LOCAL_DIR,
                          spark_log_level=SparkDefault.LOG_LEVEL,
                          memory=SparkDefault.MEMORY,
                          dep_zip=SparkDefault.DEP_ZIP):
    config += get_spark_memory_config(memory)
    session = create_spark(session_name, spark=spark, spark_local_dir=spark_local_dir,
                           config=config, packages=packages, spark_log_level=spark_log_level,
                           dep_zip=dep_zip)
    log = logging.getLogger("parquet")
    log.info("Initializing on %s", repositories)
    parquet = ParquetLoader(session, repositories)
    return parquet


def create_uast_source(args, session_name, select=HeadFiles, language_selector=None,
                       extract_uast=True):
    if args.parquet:
        parquet_loader_args = filter_kwargs(args.__dict__, create_parquet_loader)
        start_point = create_parquet_loader(session_name, **parquet_loader_args)
        root = start_point
        if extract_uast and "uast" not in [col.name for col in start_point.execute().schema]:
            raise ValueError("The parquet files do not contain UASTs.")
    else:
        engine_args = filter_kwargs(args.__dict__, create_engine)
        root = create_engine(session_name, **engine_args)
        if language_selector is None:
            language_selector = LanguageSelector(languages=args.languages)
        start_point = Ignition(root, explain=args.explain) \
            .link(select()) \
            .link(language_selector)
        if extract_uast:
            start_point = start_point.link(UastExtractor())
    return root, start_point



from typing import Iterable, Tuple

import bblfsh

from sourced.ml.algorithms.uast_to_role_id_pairs import Uast2RoleIdPairs
from sourced.ml.extractors.bags_extractor import BagsExtractor
from sourced.ml.algorithms import NoopTokenParser


class RolesAndIdsExtractor(BagsExtractor):
    """
    Extractor wrapper for Uast2RoleIdPairs algorithm.
    Note that this is an unusual BagsExtractor since it returns iterable instead of bag.

    The class did not wrap with @register_extractor because it does not produce bags as others do.
    So nobody outside code will see it or use it directly.
    For the same reason we a free to override NAMESPACE, NAME, OPTS fields with any value we want.

    TODO(zurk): Split BagsExtractor into two clases: Extractor and BagsExtractor(Extractor),
    re-inherit this class from Extractor, delete explanations from docstring.
    """
    NAMESPACE = ""
    NAME = "roles and ids"
    OPTS = {}

    def __init__(self, split_stem=False, **kwargs):
        super().__init__(**kwargs)
        self.uast2role_id_pair = Uast2RoleIdPairs(
            None, NoopTokenParser() if not split_stem else None)

    def extract(self, uast: bblfsh.Node) -> Iterable[Tuple[str, str]]:
        yield from self.uast2role_id_pair(uast)
